/**
 * @module @particle/unpack-file
 * @typicalname unpackFile
 */

const fs = require('fs');
const tar = require('tar');
const path = require('path');
const extractZip = require('extract-zip');
const throttle = require('lodash/throttle');
const progress = require('progress-stream');

/**
 * Object returned after successfully unpacking a file
 * @typedef {Object} UnpackedFile
 * @property {string} dirname - Absolute path to the directory containing your unpacked files
 */

/**
 * Unpack a tarball (`.tar.gz`, `.tgz`, etc) file and optionally report progress.
 *
 * @param {string} source - Absolute path to the tarball you want to unpack
 * @param {string} destination - Absolute path to the directory where unpacked bits will be saved
 * @param {object=} options
 * @param {function=} options.onProgress - Function to call with progess info (optional)
 * @param {number=} [options.interval=100] - How often to report progress in
 * milliseconds (optional)
 * @returns {Promise<UnpackedFile|Error>} A promise for the downloaded file or
 * an error
 *
 * @example
 *
 * const source = '/path/to/my.tar.gz';
 * const destination = '/path/to/folder';
 *
 * // unpack a tarball
 * const { dirname } = await unpackTarGz(source, destination);
 *
 * // unpack a tarball and report progress
 * const onProgress = (progress) => console.log(progress);
 * const options = { onProgress, interval: 250 };
 * const { dirname } = await unpackTarGz(source, destination, options);
 *
 */
module.exports.unpackTarGz = async (
	source,
	destination,
	{ onProgress, interval = 100 } = {}
) => {
	if (![source, destination].every(path.isAbsolute)){
		throw new Error('`source` and `destination` must use absolute paths');
	}

	const [{ size }] = await getStats(source, destination);

	return new Promise((resolve, reject) => {
		const srcStream = fs.createReadStream(source);
		const destStream = tar.x({ cwd: destination, unlink: true });

		srcStream.on('error', err => reject(err));
		destStream.on('error', err => reject(err));
		destStream.on('finish', () => resolve({ dirname: destination }));

		if (typeof onProgress !== 'function'){
			return srcStream.pipe(destStream);
		}

		const options = { time: interval, length: size };
		const progStream = progress(options, ({ length, transferred, remaining, percentage }) => {
			onProgress({ length, transferred, remaining, percentage });
		});

		return srcStream.pipe(progStream).pipe(destStream);
	});
};

/**
 * Unpack a zip (`.zip`, etc) file and optionally report progress.
 *
 * @param {string} source - Absolute path to the zip file you want to unpack
 * @param {string} destination - Absolute path to the directory where unpacked bits will be saved
 * @param {object=} options
 * @param {function=} options.onProgress - Function to call with progess info (optional)
 * @param {number=} [options.interval=100] - How often to report progress in
 * milliseconds (optional)
 * @returns {Promise<UnpackedFile|Error>} A promise for the downloaded file or
 * an error
 *
 * @example
 *
 * const source = '/path/to/my.zip';
 * const destination = '/path/to/folder';
 *
 * // unpack a tarball
 * const { dirname } = await unpackZip(source, destination);
 *
 * // unpack a tarball and report progress
 * const onProgress = (progress) => console.log(progress);
 * const options = { onProgress, interval: 250 };
 * const { dirname } = await unpackZip(source, destination, options);
 *
 */
module.exports.unpackZip = async (
	source,
	destination,
	{ onProgress, interval = 100 } = {}
) => {
	let onEntry;

	if (![source, destination].every(path.isAbsolute)){
		throw new Error('`source` and `destination` must use absolute paths');
	}

	await getStats(source, destination);

	if (typeof onProgress === 'function'){
		onEntry = throttle((entry, { entryCount, entriesRead }) => {
			onProgress({
				length: entryCount,
				transferred: entriesRead,
				remaining: entryCount - entriesRead,
				percentage: toPct(entriesRead, entryCount)
			});
		}, interval, { leading: true, trailing: false });
	}

	await extractZip(source, { dir: destination, onEntry });

	return { dirname: destination };
};


// TEST-ONLY EXPORTS //////////////////////////////////////////////////////////
module.exports.__internal__ = Symbol('internal');
module.exports[module.exports.__internal__] = {
	toPct,
	getStats,
	stat
};


// UTILS //////////////////////////////////////////////////////////////////////
function toPct(x, total){
	if (x < 1 || typeof x !== 'number' || total < 1 || typeof total !== 'number'){
		return 0;
	}
	if (x > total){
		return 100;
	}
	return (x / total) * 100;
}

function getStats(...paths){
	return Promise.all(
		paths.map(p => stat(p))
	);
}

function stat(filename){
	return new Promise((resolve, reject) => {
		fs.stat(filename, (error, stats) => {
			if (error){
				return reject(error);
			}
			return resolve(stats);
		});
	});
}

